#include "platform_win_posix.h"
#include "mp_wrap.h"

HM_EMITTER cur=0;

/* Use the following line on a Windows machine:
 #include <GL/glut.h>
 */
/* This line is for Max OSX  */
#include <GLUT/glut.h>
#include <chrono>
#include <iostream>

/*! GLUT display callback function */
void display(void);
/*! GLUT window reshape callback function */
void reshape(int, int);

void idle();


int client_wi=800;
int client_he=600;
std::chrono::system_clock::time_point currentTime;

int main(int argc, char** argv)
{
    currentTime = std::chrono::system_clock::now();

	glutInit(&argc, argv);

    glutInitWindowSize(client_wi, client_he);

	glutInitWindowPosition(100, 100);

    /* set the display mode to Red, Green, Blue and Alpha
     allocate a depth buffer
     enable double buffering
     */
    glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);

    /* create the window (and call it Lab 1) */
    glutCreateWindow("Magic Particles Wrapper for OpenGL");

    /* set the glut display callback function
     this is the function GLUT will call every time
     the window needs to be drawn
     */
    glutDisplayFunc(display);

    /* set the glut reshape callback function
     this is the function GLUT will call whenever
     the window is resized, including when it is
     first created
     */
    glutReshapeFunc(reshape);

	glutIdleFunc(idle);

    glViewport(0,0,client_wi,client_he);

    /* set the default background color to black */
    glClearColor(0,0,0,1);

	MP_Device_WRAP device(client_wi, client_he);
	device.Create();

    MP_Manager& MP=MP_Manager::GetInstance();

	MP_Platform* platform=new MP_Platform_WIN_POSIX;
    #ifdef MAGIC_3D
	MAGIC_AXIS_ENUM axis=MAGIC_pXpYnZ;
    #else
	MAGIC_AXIS_ENUM axis=MAGIC_pXnY;
    #endif

    #ifdef SHADER_WRAP
	bool filters[MAGIC_RENDER_STATE__MAX];
	for (int i=0;i<MAGIC_RENDER_STATE__MAX;i++)
		filters[i]=false;
	filters[MAGIC_RENDER_STATE_BLENDING]=true;
	filters[MAGIC_RENDER_STATE_TEXTURE]=true;
	filters[MAGIC_RENDER_STATE_ADDRESS_U]=true;
	filters[MAGIC_RENDER_STATE_ADDRESS_V]=true;
	filters[MAGIC_RENDER_STATE_ZENABLE]=true;
	filters[MAGIC_RENDER_STATE_ZWRITE]=true;
    #ifndef SHADER_ALPHATEST_WRAP
	filters[MAGIC_RENDER_STATE_ALPHATEST_INIT]=true;
	filters[MAGIC_RENDER_STATE_ALPHATEST]=true;
    #endif
    #else
	bool* filters=NULL;
    #endif

	MP.Initialization(filters, true, axis, platform, MAGIC_INTERPOLATION_ENABLE, MAGIC_CHANGE_EMITTER_ONLY, 1024, 1024, 1, 1.f, 0.1f, true);

	// eng: find of all ptc-files in folder
	// rus: поиск всех ptc-файлов в папке
	MP.LoadAllEmitters();

	MP.RefreshAtlas();

	MP.CloseFiles();

	MP.Stop();

	cur=MP.GetFirstEmitter();

    MP_Emitter* emitter=MP.GetEmitter(cur);
	emitter->SetState(MAGIC_STATE_UPDATE);

    #ifdef MAGIC_3D
    bool perspective=true;
    #else
    bool perspective=false;
    #endif
	device.SetScene3d(perspective);

    #ifndef MAGIC_3D
	// eng: locate emitters the same as editor
	// rus: расставляем эмиттеры также, как они стояли в редакторе
	HM_EMITTER hmEmitter=MP.GetFirstEmitter();
	while (hmEmitter)
	{
		Magic_CorrectEmitterPosition(hmEmitter, client_wi, client_he);
		hmEmitter=MP.GetNextEmitter(hmEmitter);
	}
	#endif

    /* enter the main event loop so that GLUT can process
     all of the window event messages
     */
    glutMainLoop();

    // eng: Delete particle manager
	// rus: Удаляем менеджер частиц
    MP.Destroy();
    device.Destroy();

    return 0;
}

/*! glut display callback function.  Every time the window needs to be drawn,
 glut will call this function.  This includes when the window size
 changes, or when another window covering part of this window is
 moved so this window is uncovered.
 */
void display()
{
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);	// eng: Black Background
    glClear(GL_COLOR_BUFFER_BIT);			// eng: Clear Screen

    MP_Manager& MP=MP_Manager::GetInstance();

	MP.Render();

    /* swap the back and front buffers so we can see what we just drew */
    glutSwapBuffers();
}

/*! glut reshape callback function.  GLUT calls this function whenever
 the window is resized, including the first time it is created.
 You can use variables to keep track the current window size.
 */
void reshape(int width, int height)
{
    /* tell OpenGL we want to display in a recangle that is the
     same size as the window
     */

    glutReshapeWindow(client_wi, client_he);
    //glViewport(0,0,width,height);
}

void idle()
{
    auto newTime = std::chrono::system_clock::now();
    auto delta = std::chrono::duration_cast<std::chrono::milliseconds>(newTime-currentTime);
    auto positionOffest = delta.count() * 0.01f;
    currentTime = newTime;

    MP_Manager& MP=MP_Manager::GetInstance();

	MP.UpdateByTimer();

	MP_Emitter* emitter=MP.GetEmitter(cur);

    MAGIC_POSITION position;
    Magic_GetEmitterPosition(emitter->GetEmitter(),&position);
    position.x += positionOffest;
    Magic_SetEmitterPosition(emitter->GetEmitter(),&position);

    if (emitter->GetState()==MAGIC_STATE_STOP)
	{
		cur=MP.GetNextEmitter(cur);
		if (!cur)
			cur=MP.GetFirstEmitter();

		emitter=MP.GetEmitter(cur);
		emitter->SetState(MAGIC_STATE_UPDATE);
	}

    display();
}
